//
//  AidlabHandler.swift
//  AidlabSDK
//
//  Created by Szymon Gęsicki on 05/02/2021.
//  Copyright © 2020 Aidlab. All rights reserved.
//

import Flutter
import AidlabSDK

public class AidlabHandler: AidlabDelegate, PlatformAidlab, InternalDelegate, InternalUpdateDelegate {
    
    private var synchronizationDelegate: AidlabSynchronizationDelegate
    private var released: Bool = false
    private var aidlabListener: PlatformAidlabDelegate?
    private var internalListener: PlatformAidlabInternalDelegate?
    private var updateListener: PlatformAidlabUpdateDelegate?
    private var connectedAidlab: IAidlab?
    private var isConnected: Bool = false
    
    init(synchronizationDelegate: AidlabSynchronizationDelegate) {
        self.synchronizationDelegate = synchronizationDelegate
    }
    
    func initial(binding: FlutterBinaryMessenger) {
        released = false
        aidlabListener = PlatformAidlabDelegate(binaryMessenger: binding)
        internalListener = PlatformAidlabInternalDelegate(binaryMessenger: binding)
        updateListener = PlatformAidlabUpdateDelegate(binaryMessenger: binding)
        PlatformAidlabSetup(binding, self)
    }
    
    func release() {
        released = true
        connectedAidlab?.disconnect()
        aidlabListener = nil
    }
    
    public func disconnect(_ error: AutoreleasingUnsafeMutablePointer<FlutterError?>) {
        connectedAidlab?.disconnect()
    }
    
    public func send(_ input: SendRequest, error: AutoreleasingUnsafeMutablePointer<FlutterError?>) {
        /// TODO throw exception
        guard let command = input.command else {
            print("Error send input")
            return
        }
        connectedAidlab?.send(command)
    }
    
    public func setECGFiltrationMethod(_ input: EcgFiltrationRequest, error: AutoreleasingUnsafeMutablePointer<FlutterError?>) {
        /// TODO throw exception
        guard let method = input.method else {
            print("Error setECGFiltrationMethod")
            return
        }
        
        let ecgFiltrationMethod: ECGFiltrationMethod = (method == 0) ? .normal : .aggressive
        
        connectedAidlab?.setECGFiltrationMethod(ecgFiltrationMethod)
    }
    
    public func startSynchronization(_ error: AutoreleasingUnsafeMutablePointer<FlutterError?>) {
        connectedAidlab?.startSynchronization()
    }
    
    public func stopSynchronization(_ error: AutoreleasingUnsafeMutablePointer<FlutterError?>) {
        connectedAidlab?.stopSynchronization()
    }
    
    public func request(forFirmwareInfo input: FirmwareInfoRequest, error: AutoreleasingUnsafeMutablePointer<FlutterError?>) {
        /// TODO throw exception
        guard let url = input.url else {
            print("Error request forFirmwareInfo")
            return
        }
        connectedAidlab?.internalAidlab.requestForFirmwareInfo(url: url)
    }
    
    public func startFirmwareUpdate(_ input: FirmwareUpdateRequest, error: AutoreleasingUnsafeMutablePointer<FlutterError?>) {
        /// TODO throw exception
        guard let url = input.url, let updatableVersion = input.updatableVersion else {
            print("Error request startFirmwareUpdate")
            return
        }
        connectedAidlab?.internalAidlab.startFirmwareUpdate(url: url, updatableVersion: updatableVersion)
    }

    public func didConnect(_ aidlab: IAidlab) {

        if released {
            aidlab.disconnect()
            return
        }
        
        isConnected = true
        connectedAidlab = aidlab
        
        connectedAidlab?.internalAidlab.internalDelegate = self
        connectedAidlab?.internalAidlab.updateInternalDelegate = self
        connectedAidlab?.aidlabSynchronizationDelegate = synchronizationDelegate
                
        aidlabListener?.didConnect(ConnectEvent().apply(closure: {
            $0.firmwareRevision = aidlab.firmwareRevision
            $0.hardwareRevision = aidlab.hardwareRevision
            $0.serialNumber = aidlab.serialNumber
            $0.address = aidlab.address?.uuidString
        }), completion: { _ in })
    }
    
    public func didDisconnect(_ aidlab: IAidlab, reason: DisconnectReason) {
        isConnected = false
        connectedAidlab?.aidlabSynchronizationDelegate = nil
        connectedAidlab = nil
        aidlabListener?.didDisconnect(DisconnectEvent().apply(closure: {
            $0.disconnectReason = NSNumber(value: reason.rawValue)
        }), completion: { _ in })
    }
    
    public func didReceiveECG(_ aidlab: IAidlab, timestamp: UInt64, values: [Float]) {
        aidlabListener?.didReceiveEcg(EcgEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.values = FlutterStandardTypedData.init(float64: Data(values: values.map { Double($0) }))
        }), completion: { _ in })
    }
    
    public func didReceiveRespiration(_ aidlab: IAidlab, timestamp: UInt64, values: [Float]) {
        aidlabListener?.didReceiveRespiration(RespirationEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.values = FlutterStandardTypedData.init(float64: Data(values: values.map { Double($0) }))
        }), completion: { _ in })
    }
    
    public func didReceiveBatteryLevel(_ aidlab: IAidlab, stateOfCharge: UInt8) {
        aidlabListener?.didReceiveBatteryLevel(BatteryLevelEvent().apply(closure: {
            $0.stateOfCharge = NSNumber(value: stateOfCharge)
        }), completion: { _ in })
    }
    
    public func didReceiveSteps(_ aidlab: IAidlab, timestamp: UInt64, value: UInt64) {
        aidlabListener?.didReceiveSteps(StepsEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.value = NSNumber(value: value)
        }), completion: { _ in })
    }
    
    public func didReceiveSkinTemperature(_ aidlab: IAidlab, timestamp: UInt64, value: Float) {
        aidlabListener?.didReceiveSkinTemperature(SkinTemperatureEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.value = NSNumber(value: value)
        }), completion: { _ in })
    }
    
    public func didReceiveAccelerometer(_ aidlab: IAidlab, timestamp: UInt64, ax: Float, ay: Float, az: Float) {
        aidlabListener?.didReceiveAccelerometer(VectorEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.x = NSNumber(value: ax)
            $0.y = NSNumber(value: ay)
            $0.z = NSNumber(value: az)
        }), completion: { _ in })
    }
    
    public func didReceiveGyroscope(_ aidlab: IAidlab, timestamp: UInt64, qx: Float, qy: Float, qz: Float) {
        aidlabListener?.didReceiveGyroscope(VectorEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.x = NSNumber(value: qx)
            $0.y = NSNumber(value: qy)
            $0.z = NSNumber(value: qz)
        }), completion: { _ in })
    }
    
    public func didReceiveMagnetometer(_ aidlab: IAidlab, timestamp: UInt64, mx: Float, my: Float, mz: Float) {
        aidlabListener?.didReceiveMagnetometer(VectorEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.x = NSNumber(value: mx)
            $0.y = NSNumber(value: my)
            $0.z = NSNumber(value: mz)
        }), completion: { _ in })
    }
    
    public func didReceiveQuaternion(_ aidlab: IAidlab, timestamp: UInt64, qw: Float, qx: Float, qy: Float, qz: Float) {
        aidlabListener?.didReceiveQuaternion(QuaternionEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.qw = NSNumber(value: qw)
            $0.qx = NSNumber(value: qx)
            $0.qy = NSNumber(value: qy)
            $0.qz = NSNumber(value: qz)
        }), completion: { _ in })
    }
    
    public func didReceiveOrientation(_ aidlab: IAidlab, timestamp: UInt64, roll: Float, pitch: Float, yaw: Float) {
        aidlabListener?.didReceiveOrientation(OrientationEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.roll = NSNumber(value: roll)
            $0.pitch = NSNumber(value: pitch)
            $0.yaw = NSNumber(value: yaw)
        }), completion: { _ in })
    }
    
    public func didReceiveBodyPosition(_ aidlab: IAidlab, timestamp: UInt64, bodyPosition: BodyPosition) {
        aidlabListener?.didReceiveBodyPosition(BodyPositionEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.bodyPosition = NSNumber(value: bodyPosition.rawValue)
        }), completion: { _ in })
    }
    
    public func didReceiveHeartRate(_ aidlab: IAidlab, timestamp: UInt64, heartRate: Int32) {
        aidlabListener?.didReceiveHeartRate(HeartRateEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.heartRate = NSNumber(value: heartRate)
        }), completion: { _ in })
    }
    
    public func didReceiveHrv(_ aidlab: IAidlab, timestamp: UInt64, hrv: Int32) {
        aidlabListener?.didReceiveHrv(HrvEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.hrv = NSNumber(value: hrv)
        }), completion: { _ in })
    }
    
    public func didReceiveRespirationRate(_ aidlab: IAidlab, timestamp: UInt64, value: UInt32) {
        aidlabListener?.didReceiveRespirationRate(RespirationRateEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.value = NSNumber(value: value)
        }), completion: { _ in })
    }
    
    public func didReceiveSoundVolume(_ aidlab: IAidlab, timestamp: UInt64, soundVolume: UInt16) {
        aidlabListener?.didReceiveSoundVolume(SoundVolumeEvent().apply(closure: {
            $0.timestamp = NSNumber(value: soundVolume)
            $0.soundVolume = NSNumber(value: soundVolume)
        }), completion: { _ in })
    }
    
    public func didDetectUserEvent(_ aidlab: IAidlab, timestamp: UInt64) {
        aidlabListener?.didDetect(UserEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
        }), completion:  { _ in })
    }
    
    public func didReceivePressure(_ aidlab: IAidlab, timestamp: UInt64, values: [Int32]) {
        internalListener?.didReceivePressure(PressureEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.values = FlutterStandardTypedData.init(int32: Data(values: values.map { Int32($0) }))
        }), completion: { _ in })
    }
    
    public func didDetect(_ aidlab: IAidlab, exercise: Exercise) {
        aidlabListener?.didDetectExercise(ExerciseEvent().apply(closure: {
            $0.exercise = NSNumber(value: exercise.rawValue)
        }), completion: { _ in })
    }
    
    public func didDetect(_ aidlab: IAidlab, timestamp: UInt64, activity: ActivityType) {
        aidlabListener?.didReceiveActivity(ActivityEvent().apply(closure: {
            $0.timestamp = NSNumber(value: timestamp)
            $0.activityType = NSNumber(value: UInt64(activity.rawValue))
        }), completion: { _ in })
    }
    
    public func didReceiveError(_ aidlab: IAidlab, error: Error) {
        aidlabListener?.didReceiveError(ErrorEvent().apply(closure: {
            $0.error = error.localizedDescription
        }), completion: { _ in })
    }
    
    public func didReceiveRSSI(_ aidlab: IAidlab, rssi: Int32) {
        /// TODO !
    }
    
    public func wearStateDidChange(_ aidlab: IAidlab, state: WearState) {
        aidlabListener?.wearStateDidChange(WearStateEvent().apply(closure: {
            $0.wearState = NSNumber(value: state.rawValue)
        }), completion: { _ in })
    }
    
    public func pressureWearStateDidChange(_ aidlab: IAidlab, state: WearState) {
        internalListener?.pressureWearStateDidChange(PressureWearStateEvent().apply(closure: {
            $0.wearState = NSNumber(value: state.rawValue)
        }), completion: { _ in })
    }
    
    public func didReceiveCommand(_ aidlab: IAidlab) {
        aidlabListener?.didReceiveCommand(EmptyEvent(), completion: { _ in })
    }
    
    public func didReceiveMessage(_ aidlab: IAidlab, process: String, message: String) {
        aidlabListener?.didReceiveMessage(MessageEvent().apply(closure: {
            $0.process = process
            $0.message = message
        }), completion: { _ in })
    }
    
    public func didReceiveFirmwareInformation(updateAvailable: Bool, updatableVersion: String) {
        internalListener?.didReceiveFirmwareInformation(FirmwareInformationEvent().apply(closure: {
            $0.updateAvailable = NSNumber(value: updateAvailable)
            $0.updatableVersion = updatableVersion
        }), completion: { _ in })
    }
    
    public func didReceiveSoundFeatures(_ aidlab: IAidlab, values: [Float]) {
        aidlabListener?.didReceiveSoundFeatures(SoundFeaturesEvent().apply(closure: {
            $0.values = FlutterStandardTypedData.init(float64: Data(values: values.map { Double($0) }))
        }), completion: { _ in })
    }
    
    public func didReceiveUpdateError(error: UpdateError) {
        updateListener?.didReceiveUpdateError(UpdateEvent().apply(closure: {
            $0.error = NSNumber(value: error.rawValue)
        }), completion: {_ in })
    }
    
    public func didReceiveUpdateProgress(progress: UInt8) {
        updateListener?.didReceive(UpdateProgress().apply(closure: {
            $0.progress = NSNumber(value: progress)
        }), completion: { _ in })
    }
    
    public func didFinishUpdate() {
        updateListener?.didFinishUpdate(EmptyEvent(), completion: { _ in })
    }
}
